# AddOn Tile Proc (c) 2014 Atom

import bpy
import os,sys,colorsys

import mathutils
from mathutils import Vector, Matrix

import random
from random import randint, uniform

#####################################################################
# Globals.
#####################################################################
# Global busy flag. Try to avoid events if we are already busy.
isBusy = False
isRendering = False
lastFrameUpdated = 0.0
frameChangeCount = 0
      
# Objects are managed by name prefix. Customize here...
OBJECT_PREFIX = "tileproc_"      	#For master control object naming.
TILEPROC_OB_PREFIX = "tp_"       	#For managed object naming.
TILEPROC_MESH_PREFIX = "me_tp_"  	#For managed datablock naming.
MATERIAL_PREFIX = "mat_tp_"			#For managed material naming.
ENTRY_NAME = "tile"					#For new list entries.
ITEM_NAME = "item"					#For item type entries.
                
MAX_NAME_SIZE = 52
GLOBAL_ZERO_PADDING = 3             # The number of zeros to pad strings with when converting INTs to STRINGs.
DELIMITER = ","

#####################################################################
# Simple debug message control.
#####################################################################
SHOW_MESSAGES = True
MSG_PREFIX = "tileproc} "
def to_console(passedItem):
    if SHOW_MESSAGES == True:
        if len(passedItem) == 0:
            print("")
        else:
            s = str(passedItem)
            print(MSG_PREFIX + s)
#####################################################################
# Memory management
#####################################################################
def removeObjectFromMemory (passedName):
    to_console("removeObjectFromMemory:[%s]." % passedName)
    # Extra test because this can crash Blender if not done correctly.
    result = False
    ob = bpy.data.objects.get(passedName)
    if ob != None:
        if ob.users == 0:
            try:
                ob.user_clear()
                can_continue = True
            except:
                can_continue = False
            
            if can_continue == True:
                try:
                    bpy.data.objects.remove(ob)
                    result = True
                    to_console("removeObjectFromMemory: OB [" + passedName + "] removed from memory.")
                except:
                    result = False
                    to_console("removeObjectFromMemory: FAILED to remove OB [" + passedName + "] from memory.")
            else:
                # Unable to clear users, something is holding a reference to it.
                # Can't risk removing. Favor leaving it in memory instead of risking a crash.
                to_console("removeObjectFromMemory: Unable to clear users for OBJECT, something is holding a reference to it.")
                result = False
        else:
            to_console ("removeObjectFromMemory: Unable to remove OBJECT because it still has [%s] users." % ob.users) 
    else:
        # We could not fetch it, it does not exist in memory, essentially removed.
        to_console("We could not fetch OB [%s], it does not exist in memory, essentially removed." % passedName)
        result = True
    return result
  
def removeCurveFromMemory (passedName):
    # Extra test because this can crash Blender if not done correctly.
    result = False
    curve = bpy.data.curves.get(passedName)
    if curve != None:
        if curve.users == 0:
            try:
                curve.user_clear()
                can_continue = True
            except:
                can_continue = False
            
            if can_continue == True:
                try:
                    bpy.data.curves.remove(curve)
                    result = True
                    #to_console("removeCurveFromMemory: CURVE [" + passedName + "] removed from memory.")
                except:
                    result = False
                    to_console("removeCurveFromMemory: FAILED to remove [" + passedName + "] from memory.")
            else:
                # Unable to clear users, something is holding a reference to it.
                # Can't risk removing. Favor leaving it in memory instead of risking a crash.
                to_console("removeCurveFromMemory: Unable to clear users for MESH, something is holding a reference to it.")
                result = False
        else:
            to_console ("removeCurveFromMemory: Unable to remove CURVE because it still has [" + str(curve.users) + "] users.")
    else:
        # We could not fetch it, it does not exist in memory, essentially removed.
        result = True
    return result
    
def eraseAllKeyframes(passedOB = None):
    if passedOB != None:
        ad = passedOB.animation_data
        if ad != None:
            passedOB.animation_data_clear()
#####################################################################
# Name management
##################################################################### 
def returnAllObjectNames (scene):
    # NOTE: This returns all object names in the passed scene.
    result = []
    for ob in scene.objects:
        result.append(ob.name)
    return result 

def returnObjectNamesLike(passedScene, passedName):
    # Return objects named like our passedName in the passed scene.
    result = []
    isLike = passedName
    l = len(isLike)
    all_obs = returnAllObjectNames(passedScene)
    for name in all_obs:
        candidate = name[0:l]
        if isLike == candidate:
            result.append(name)
    return result

def returnNameForNumber(passedFrame):
    frame_number = str(passedFrame)
    post_fix = frame_number.zfill(GLOBAL_ZERO_PADDING)
    return post_fix

def returnNameDroppedPrefix(ob):
    # Use only right half of name, i.e. discard xxxx_ prefix.
    lst_name = ob.name.split('_')
    return lst_name[1]

def returnTileName(passedName, passedX, passedY):
    return "%s_%sx%s" % (passedName,returnNameForNumber(passedX),returnNameForNumber(passedY))
    
def returnItemName(passedName, passedX, passedY, passedCount):
    return "%s_%sx%s-%s" % (passedName,returnNameForNumber(passedX),returnNameForNumber(passedY),returnNameForNumber(passedCount))

def returnCurveName(passedName, passedX, passedY):
    return "%s_%sx%s" % (passedName,returnNameForNumber(passedX),returnNameForNumber(passedY))

def returnIndexedName(passedName, passedIndex):
    return "%s_%s" % (passedName,returnNameForNumber(passedIndex))

############################################################################
# Tile code.
############################################################################        
def returnTileNeighborTypes(grid, x, y):
    top = -1
    bottom = -1
    left = -1
    right = -1
    w = len(grid)
    h = len(grid[0])
    x1 = x-1
    y1 = y-1
    x2 = x+1
    y2 = y+1
    if y1 >= 0:
        top = grid[x][y1]
    if y2 < h:
        bottom = grid[x][y2]
    if x1 >= 0:
        left = grid[x1][y]
    if x2 < w:
        right = grid[x2][y]

    return top, bottom, left, right
    
def returnMapNeighborTypes(grid, x, y):
    top = "#"		# Solid wall symbol.
    bottom = "#"
    left = "#"
    right = "#"
    w = len(grid)
    h = len(grid[0])
    x1 = x-1
    y1 = y-1
    x2 = x+1
    y2 = y+1
    if y1 >= 0:
        top = grid[x][y1]
    if y2 < h:
        bottom = grid[x][y2]
    if x1 >= 0:
        left = grid[x1][y]
    if x2 < w:
        right = grid[x2][y]

    return top, bottom, left, right 
       
def returnSelfDeterminedByNeighbors(top, bottom, left, right):
    result = None
    # A Quad intersection.
    if top == "." and bottom == "." and left == "." and right == ".": result = "XX"
    
    # T intersections.
    if top == "#" and bottom == "." and left == "." and right == ".": result = "TT"
    if top == "." and bottom == "#" and left == "." and right == ".": result = "TB"
    if top == "." and bottom == "." and left == "#" and right == ".": result = "TL"
    if top == "." and bottom == "." and left == "." and right == "#": result = "TR"
    
    # Corner types.
    if top == "#" and bottom == "." and left == "#" and right == ".": result = "BR"
    if top == "#" and bottom == "." and left == "." and right == "#": result = "BL"
    if top == "." and bottom == "#" and left == "#" and right == ".": result = "RT"
    if top == "." and bottom == "#" and left == "." and right == "#": result = "LT"
    
    # Vertical, horizontal.
    if top == "#" and bottom == "#" and left == "." and right == ".": result = "=="
    if top == "." and bottom == "." and left == "#" and right == "#": result = "||"
    
    # Single types.
    if top == "." and bottom == "#" and left == "#" and right == "#": result = "ST"
    if top == "#" and bottom == "." and left == "#" and right == "#": result = "SB"
    if top == "#" and bottom == "#" and left == "." and right == "#": result = "SL"
    if top == "#" and bottom == "#" and left == "#" and right == ".": result = "SR"
    return result
    
############################################################################
# Auto drive code.
############################################################################
def returnFollowPathConstraint(passedName, passedObject, passedTarget, passedValue = 1.0):
    # Create or fetch a follow path constraint and return it.
    #try:
    # See if our constraint exists.
    con_name = passedName
    con_name = con_name[:MAX_NAME_SIZE]
    my_constraint = passedObject.constraints.get(con_name)    #fetchIfConstraint(con_name, passedObject)
    if my_constraint == None:
        # Create a new constraint.
        my_constraint = passedObject.constraints.new("FOLLOW_PATH")
        my_constraint.name = con_name

    my_constraint.use_curve_follow = True
    my_constraint.up_axis = 'UP_Z'   
    my_constraint.forward_axis =  'TRACK_NEGATIVE_Y'
    my_constraint.target = passedTarget
    my_constraint.influence = passedValue
    my_constraint.use_fixed_location = True 
    my_constraint.offset_factor = 0.0   # Animate this value to move along the path.
    result = my_constraint
    #except:
    #    print("returnFollowPathConstraint: Problem assigning constraint [%s]." % passedName)
    #    result = None
    return result

def curveFromEndPoints(curveName = "lerp", v1=Vector((0,0,0)),v2=Vector((0,0,1)),subdiv = 6, noise=0.0, threshold=0.0):
    v_sliced = [v1] + [v1.lerp(v2, i/subdiv) for i in range(1, subdiv + 1)]
    
    # Now make a curve out of the new vertices.
    curve_datablock_name = "cu_%s" % curveName
    cu_datablock = bpy.data.curves.get(curve_datablock_name)
    if cu_datablock == None:
        cu_datablock = bpy.data.curves.new(curve_datablock_name,'CURVE')    # Create a new curve.
        cu_datablock.dimensions = '3D'
    else:
        #Reuse this existing curve, but we must remove all of the current splines.
        for old_spline in cu_datablock.splines:
            try:
                cu_datablock.splines.remove(old_spline)
            except:
                print("curveFromEndPoints: ERROR trying to remove old_spline")
    
    curve_name = curveName
    ob_curve = bpy.data.objects.get(curve_name)
    if ob_curve == None:
        try:    
            ob_curve = bpy.data.objects.new(curve_name,cu_datablock)                    # Create a new curve type object.
            bpy.context.scene.objects.link(ob_curve)                                    # Link it to the scene.
        except:
            print("curveFromEndPoints: ERROR, failed to create tracer curve to [%s]." % curve_name)
    else:
        #print("WARNING: [" + curve_name + "] already exists.")
        pass
        
    # Populate the points.
    spline = cu_datablock.splines.new('BEZIER')               # Add an empty spline to the curve.
    spline.bezier_points.add(len(v_sliced)-1)                   # Pre-create the number of points in this spline based upon passed value.
    for t,co in enumerate(v_sliced):
        canRandomize = True
        if t == 0: canRandomize = False
        if t == (len(v_sliced)-1): canRandomize = False
        if canRandomize == True:
            # Only randomize points inbetween the end points.
            if random.random() > threshold:
                #p.radius = random.random()
                for i in range(3):
                    co[i] += noise*(random.random()*2-1)
                     
        p = spline.bezier_points[t]
        p.co = co 
        p.tilt = 0.0
        #p.radius = 0.0
        p.handle_right_type='AUTO'
        p.handle_left_type='AUTO'
    return ob_curve

def returnActionFromObject(passedOb):
    try:
        action_name = passedOb.animation_data.action.name
    except:
        action_name = ""
    try:
        action = bpy.data.actions[action_name]
    except:
        action = None
    return action
    
############################################################################
# Color code.
############################################################################
def returnRandomGauss():
    return random.gauss(5,1)
    
def randFloat(a, b):
    return random.random()*(b-a)+a

def returnRandomColor (color_base = None, color_match = None):
    if color_base == None:
        # Ugly random color.
        return (randFloat(0.2,1.0),randFloat(0.2,1.0),randFloat(0.2,1.0), 1.0)
    else:
        if color_match == None:
            # Random color based upon base color hue.
            h, s, v = colorsys.rgb_to_hsv(color_base[0],color_base[1],color_base[2])
            # Use the HUE from color base to determine this unique random color.
            return (returnRndRGBFromHue (h))
        else:
            # Use hue from the color base and the value from the color_match.
            hb, sb, vb = colorsys.rgb_to_hsv(color_base[0],color_base[1],color_base[2])
            hm, sm, vm = colorsys.rgb_to_hsv(color_match[0],color_match[1],color_match[2])
            s = uniform(0.4, 1.0)
            r, g, b = colorsys.hsv_to_rgb(hb, s, vm)
            return r,g,b, 1.0
        
def returnNumberOfColors(count):
    N = count
    HSV_tuples = [(x*1.0/N, 0.5, 0.5) for x in range(N)]
    RGB_tuples = map(lambda x: colorsys.hsv_to_rgb(*x), HSV_tuples)
    return RGB_tuples

def returnRndRGBFromHue (h):
    #h = uniform(0.25, 0.38) # Select random green'ish hue from hue wheel
    s = uniform(0.4, 1.0)
    v = uniform(0.2, 0.9)
    r, g, b = colorsys.hsv_to_rgb(h, s, v)
    return r,g,b, 1.0

############################################################################
# BEGIN:
# McKracken Curve Tools code.
############################################################################
#cuadratic bezier value
def quad(p, t):
    return p[0]*(1.0-t)**2.0 + 2.0*t*p[1]*(1.0-t) + p[2]*t**2.0

#cubic bezier value
def cubic(p, t):
    return p[0]*(1.0-t)**3.0 + 3.0*p[1]*t*(1.0-t)**2.0 + 3.0*p[2]*(t**2.0)*(1.0-t) + p[3]*t**3.0

#gets a bezier segment's control points on global coordinates
def getbezpoints(spl, mt, seg=0):
    points = spl.bezier_points
    p0 = points[seg].co * mt
    p1 = points[seg].handle_right * mt
    p2 = points[seg+1].handle_left * mt
    p3 = points[seg+1].co * mt  
    return p0, p1, p2, p3

#gets nurbs polygon control points on global coordinates
def getnurbspoints(spl, mw):
    pts = []
    ws = []
    for p in spl.points:
        v = Vector(p.co[0:3])*mw
        pts.append(v)
        ws.append(p.weight)
    return pts , ws

#calculates a global parameter t along all control points
#t=0 begining of the curve
#t=1 ending of the curve
def calct(obj, t):
    
    spl=None
    mw = obj.matrix_world
    if obj.data.splines.active==None:
        if len(obj.data.splines)>0:
            spl=obj.data.splines[0]
    else:
        spl = obj.data.splines.active
    
    if spl==None:
        return False
    
    if spl.type=="BEZIER":
        points = spl.bezier_points
        nsegs = len(points)-1
        
        d = 1.0/nsegs
        seg = int(t/d)
        t1 = t/d - int(t/d)
        
        if t==1:
            seg-=1
            t1 = 1.0
        
        p = getbezpoints(spl,mw, seg)
        
        coord = cubic(p, t1)
        
        return coord

    elif spl.type=="NURBS":
        data = getnurbspoints(spl, mw)
        pts = data[0]
        ws = data[1]
        order = spl.order_u
        n = len(pts)
        ctype = spl.use_endpoint_u
        kv = knots(n, order, ctype)
        
        coord = C(t, order-1, pts, ws, kv)
        
        return coord

#length of the curve
def arclength(objs):
    length = 0.0
    
    for obj in objs:
        if obj.type=="CURVE":
            prec = 100   #precision
            inc = 1/prec #increments
            
            ### TODO: set a custom precision value depending the number of curve points
            #that way it can gain on accuracy in less operations.
            
            #subdivide the curve in 100 lines and sum its magnitudes
            for i in range(0, prec):
                ti = i*inc 
                tf = (i+1)*inc 
                a = calct(obj, ti) 
                b = calct(obj, tf)
                r = (b-a).magnitude 
                length+=r
            
    return length
############################################################################
# END:
# Mackraken Curve Tools code.
############################################################################
